"""
Draft Sport Python
Competition Constants Module
author: hugh@blinkybeach.com
"""
from draft_sport.fantasy.competition.competition import Competition


class CompetitionConstants:

    DRAFT_RUGBY_2020 = Competition(
        public_id='sru_2020_2',
        name='Draft Rugby 2020'
    )

    DRAFT_RUGBY_2020_PRE_COVID = Competition(
        public_id='sru_2020_1',
        name='Draft Rugby 2020 (Pre-Covid)'
    )



"""
Draft Sport
Position Requirement Module
author: hugh@blinkybeach.com
"""
from nozomi import Immutable, Decodable
from typing import Type, TypeVar, Any

T = TypeVar('T', bound='PositionRequirement')


class PositionRequirement(Decodable):

    def __init__(
        self,
        count: int,
        position_name: str
    ) -> None:

        self._count = count
        self._position_name = position_name

        return

    count: int = Immutable(lambda s: s._count)
    position_name: str = Immutable(lambda s: s._position_name)

    @classmethod
    def decode(cls: Type[T], data: Any) -> T:
        return cls(
            count=data['count'],
            position_name=data['position_name']
        )



"""
Draft Sport Python
Fantasy Team Class
author: hugh@blinkybeach.com
"""
from nozomi import Immutable


class Team:

    def __init__(
        self,
        name: str
    ) -> None:

        self._name = name

        return

    name = Immutable(lambda s: s._name)
    lowercase_name = Immutable(lambda s: s._name.lower())



"""
Draft Sport Python
Team Constants Module
author: hugh@blinkybeach.com
"""
from draft_sport.fantasy.team.team import Team


class TeamConstants:

    SUPER_RUGBY = (
        Team('Blues'),
        Team('Brumbies'),
        Team('Bulls'),
        Team('Chiefs'),
        Team('Crusaders'),
        Team('Highlanders'),
        Team('Hurricanes'),
        Team('Jaguares'),
        Team('Lions'),
        Team('NSW Waratahs'),
        Team('Rebels'),
        Team('Reds'),
        Team('Sharks'),
        Team('Stormers'),
        Team('Sunwolves')
    )



"""
Draft Sport Python
Composition Module
author: hugh@blinkybeach.com
"""
from nozomi import Immutable, Decodable
from typing import TypeVar, Type, Any, List
from draft_sport.leagues.position_requirement import PositionRequirement
from draft_sport.leagues.category_requirement import CategoryRequirement

T = TypeVar('T', bound='Composition')


class Composition(Decodable):

    def __init__(
        self,
        position_requirements: List[PositionRequirement],
        category_requirements: List[CategoryRequirement]
    ) -> None:

        self._position_requirements = position_requirements
        self._category_requirements = category_requirements

        return

    position_requirements = Immutable(
        lambda s: s._position_requirements
    )
    category_requirements = Immutable(
        lambda s: s._category_requirements
    )
    unique_positions: List[str] = Immutable(
        lambda s: list(set(r.position for r in s._position_requirements))
    )

    @classmethod
    def decode(cls: Type[T], data: Any) -> T:
        return cls(
            position_requirements=PositionRequirement.decode_many(
                data['position_requirements']
            ),
            category_requirements=CategoryRequirement.decode_many(
                data['category_requirements']
            )
        )



"""
Draft Sport Python
Universe Position Constants Module
author: hugh@blinkybeach.com
"""
from draft_sport.universe.position.category import PositionCategory


class CategoryConstants:

    SUPER_RUGBY_BACK = PositionCategory(1, 'back')
    SUPER_RUGBY_FORWARD = PositionCategory(2, 'forward')

    SUPER_RUGBY = (
        SUPER_RUGBY_FORWARD,
        SUPER_RUGBY_BACK
    )



"""
Draft Sport
Position Requirement Module
author: hugh@blinkybeach.com
"""
from draft_sport.universe.position.category import PositionCategory
from nozomi import Immutable, Decodable
from typing import Type, TypeVar, Any

T = TypeVar('T', bound='PositionRequirement')


class CategoryRequirement(Decodable):

    def __init__(
        self,
        count: int,
        category: PositionCategory
    ) -> None:

        self._count = count
        self._category = category

        return

    count: int = Immutable(lambda s: s._count)
    category: PositionCategory = Immutable(lambda s: s._category)

    @classmethod
    def decode(cls: Type[T], data: Any) -> T:
        return cls(
            count=data['count'],
            category=PositionCategory.decode(data['category'])
        )



"""
Draft Sport Python
Fantasy Competition Module
author: hugh@blinkybeach.com
"""
from typing import TypeVar, Type, Any
from nozomi import Decodable

T = TypeVar('T', bound='Competition')


class Competition(Decodable):

    def __init__(
        self,
        public_id: str,
        name: str
    ) -> None:

        self._public_id = public_id
        self._name = name

        return

    @classmethod
    def decode(cls: Type[T], data: Any) -> T:
        return cls(
            public_id=data['public_id'],
            name=data['name']
        )



"""
Draft Sport Python
Fantasy Player Score Module
Author: hugh@blinkybeach.com
"""
from nozomi import Decodable, HTTPMethod, URLParameter, URLParameters
from nozomi import Immutable, ApiRequest, RequestCredentials, Encodable
from draft_sport.fantasy.player.profile import Profile
from draft_sport.fantasy.season.season import Season
from typing import Type, TypeVar, Any, Optional, List
from draft_sport.ancillary.configuration import Configuration
from draft_sport.fantasy.scores.player.order_by import OrderBy
from draft_sport.fantasy.scores.player.points import Points
from draft_sport.universe.position.category import PositionCategory
from draft_sport.data.order import Order
from typing import Dict

T = TypeVar('T', bound='Player')


class ScoreCard(Decodable, Encodable):

    _PATH = '/fantasy/score-card'
    _LIST_PATH = '/fantasy/score-card/list'

    def __init__(
        self,
        profile: Profile,
        season_id: str,
        limit: int,
        offset: int,
        query_count: int,
        sequence: int,
        requesting_agent_id: str,
        points: Points,
        query_time: int
    ) -> None:

        self._profile = profile
        self._season_id = season_id
        self._limit = limit
        self._offset = offset
        self._query_count = query_count
        self._sequence = sequence
        self._requesting_agent_id = requesting_agent_id
        self._points = points
        self._query_time = query_time

        return

    public_id = Immutable(lambda s: s._profile.public_id)

    profile = Immutable(lambda s: s._profile)
    season_id = Immutable(lambda s: s._season_id)
    limit = Immutable(lambda s: s._limit)
    offset = Immutable(lambda s: s._offset)
    query_count = Immutable(lambda s: s._query_count)
    sequence = Immutable(lambda s: s._sequence)
    requesting_agent_id = Immutable(lambda s: s._requesting_agent_id)
    points = Immutable(lambda s: s._points)
    query_time = Immutable(lambda s: s._query_time)

    query_time_seconds = Immutable(
        lambda s: str(int(s._query_time) / 1000000)
    )

    def has_position_with_name(self, name: str) -> bool:
        """Return True if this Player has the named position"""
        return self._profile.position_name == name

    def has_position_in_category(self, category: PositionCategory) -> bool:
        """Return True if this Player has a position in the supplied category"""
        return category in self._profile.position.categories

    def encode(self) -> Dict[str, Any]:
        return {
            'player': self._profile.encode(),
            'limit': self._limit,
            'offset': self._offset,
            'sequence': self._sequence,
            'requesting_agent_id': self._requesting_agent_id,
            'points': self._points.encode(),
            'query_time': self._query_time,
            'query_count': self._query_coun
        }

    @classmethod
    def decode(cls: Type[T], data: Any) -> T:
        return cls(
            profile=Profile.decode(data['player']),
            season_id=data['fantasy_season_id'],
            limit=data['limit'],
            offset=data['offset'],
            query_count=data['query_count'],
            sequence=data['sequence'],
            requesting_agent_id=data['requesting_agent_id'],
            points=Points.decode(data['points']),
            query_time=data['query_time']
        )

    @classmethod
    def retrieve(
        cls: Type[T],
        public_id: str,
        season: Season,
        credentials: Optional[RequestCredentials],
        configuration: Optional[Configuration]
    ) -> Optional[T]:

        parameters = URLParameters([
            URLParameter('public_id', public_id),
            URLParameter('season', season.public_id)
        ])

        request = ApiRequest(
            path=cls._PATH,
            method=HTTPMethod.GET,
            data=None,
            url_parameters=parameters,
            credentials=credentials,
            configuration=configuration
        )

        return cls.optionally_decode(request.response_data)

    @classmethod
    def retrieve_many(
        cls: Type[T],
        season: Season,
        offset: int,
        limit: int,
        configuration: Configuration,
        order_by: OrderBy = OrderBy.TOTAL_SEASON_POINTS,
        order: Order = Order.DESCENDING,
        name_fragment: Optional[str] = None,
        credentials: Optional[RequestCredentials] = None
    ) -> List[T]:

        parameters = URLParameters([
            URLParameter('season', season.public_id),
            URLParameter('offset', offset),
            URLParameter('limit', limit),
            URLParameter('order_by', order_by.value),
            URLParameter('order', order.value)
        ])

        if name_fragment is not None:
            parameters.append(URLParameter('fragment', name_fragment))

        request = ApiRequest(
            path=cls._LIST_PATH,
            method=HTTPMethod.GET,
            data=None,
            url_parameters=parameters,
            credentials=credentials,
            configuration=configuration
        )

        return cls.optionally_decode_many(
            request.response_data,
            default_to_empty_list=True
        )



"""
Draft Sport Python
Season Constants Module
author: hugh@blinkybeach.com
"""
from draft_sport.fantasy.season.season import Season
from draft_sport.fantasy.competition.constants import CompetitionConstants
from nozomi import NozomiDate


class SeasonConstants:

    DRAFT_RUGBY_2020 = Season(
        public_id='draft_rugby_2020',
        name='Draft Rugby 2020',
        start_date=NozomiDate(year=2020, month=7, day=4),
        competition=CompetitionConstants.DRAFT_RUGBY_2020
    )

    DRAFT_RUGBY_2020_PRE_COVID = Season(
        public_id='draft_rugby_2020_1',
        name='Draft Rugby 2020 (Pre-Covid)',
        start_date=NozomiDate(year=2020, month=2, day=5),
        competition=CompetitionConstants.DRAFT_RUGBY_2020_PRE_COVID
    )



"""
Draft Sport API
Filled Requirement Module
author: hugh@blinkybeach.com
"""
from nozomi import Immutable
from typing import List, Union
from draft_sport.leagues.pick import Pick
from draft_sport.leagues.position_requirement import PositionRequirement
from draft_sport.leagues.category_requirement import CategoryRequirement

Requirement = Union[PositionRequirement, CategoryRequirement]


class FilledRequirement:
    """
    Utility class coralling Requirement & Player data. Useful for client
    presentation purposes. There is no equivalent object in the
    Draft Sport API
    """

    def __init__(
        self,
        requirement: Requirement,
        picks: List[Pick]
    ) -> None:

        self._picks = picks
        self._requirement = requirement

        return

    picks = Immutable(lambda s: s._picks)
    name = Immutable(lambda s: s._compute_requirement_name())
    position_name = Immutable(lambda s: s._requirement.position_name)
    count = Immutable(lambda s: s._requirement.count)

    is_position: bool = Immutable(
        lambda s: isinstance(s._requirement, PositionRequirement)
    )
    is_category: bool = Immutable(
        lambda s: isinstance(s._requirement, CategoryRequirement)
    )
    is_plural: bool = Immutable(lambda s: s.count > 1)
    number_filled: int = Immutable(lambda s: len(s.picks))
    empty: bool = Immutable(lambda s: len(s.picks) < 1)
    satisfied: bool = Immutable(lambda s: len(s.picks) == s.count)
    not_satisfied: bool = Immutable(lambda s: not s.satisfied)

    def _compute_requirement_name(self) -> str:

        if isinstance(self._requirement, CategoryRequirement):
            name = self._requirement.category.name
            return name[0].upper() + name[1:].lower()

        if isinstance(self._requirement, PositionRequirement):
            return self._requirement.position_name

        raise TypeError



"""
Draft Sport Python
Fantasy Position Constants Module
author: hugh@blinkybeach.com
"""
from draft_sport.fantasy.position.position import Position
from draft_sport.universe.position.constants import CategoryConstants

_FORWARD_CAT = [CategoryConstants.SUPER_RUGBY_FORWARD]
_BACK_CAT = [CategoryConstants.SUPER_RUGBY_BACK]


class PositionConstants:

    SUPER_RUGBY_PROP = Position(1, 'Prop', _FORWARD_CAT)
    SUPER_RUGBY_HOOKER = Position(2, 'Hooker', _FORWARD_CAT)
    SUPER_RUGBY_LOCK = Position(3, 'Lock', _FORWARD_CAT)
    SUPER_RUGBY_BACK_ROW = Position(4, 'Back Row', _FORWARD_CAT)
    SUPER_RUGBY_SCRUMHALF = Position(5, 'Scrumhalf', _BACK_CAT)
    SUPER_RUGBY_FLYHALF = Position(6, 'Flyhalf', _BACK_CAT)
    SUPER_RUGBY_OUTSIDE_BACK = Position(7, 'Outside Back', _BACK_CAT)
    SUPER_RUGBY_CENTRE = Position(8, 'Centre', _BACK_CAT)

    SUPER_RUGBY = (
        SUPER_RUGBY_PROP,
        SUPER_RUGBY_HOOKER,
        SUPER_RUGBY_LOCK,
        SUPER_RUGBY_BACK_ROW,
        SUPER_RUGBY_SCRUMHALF,
        SUPER_RUGBY_FLYHALF,
        SUPER_RUGBY_OUTSIDE_BACK,
        SUPER_RUGBY_CENTRE
    )



"""
Draft Sport Python
Fantasy Position Module
author: hugh@blinkybeach.com
"""
from nozomi import Decodable, Immutable
from draft_sport.universe.position.category import PositionCategory
from typing import TypeVar, Type, Any, List

T = TypeVar('T', bound='Position')


class Position(Decodable):

    def __init__(
        self,
        indexid: int,
        name: str,
        categories: List[PositionCategory]
    ) -> None:

        self._indexid = indexid
        self._name = name
        self._categories = categories

        return

    name = Immutable(lambda s: s._name)
    lowercase_name = Immutable(lambda s: s._name.lower())

    sole_category = Immutable(
        lambda s: s._categories[0] if len(s._categories) == 1 else None
    )

    @classmethod
    def decode(cls: Type[T], data: Any) -> T:
        return cls(
            indexid=data['indexid'],
            name=data['name'],
            categories=PositionCategory.optionally_decode_many(
                data['categories'],
                default_to_empty_list=True
            )
        )

    def __eq__(self, other) -> bool:
        if not isinstance(other, Position):
            return False
        if self._indexid == other._indexid:
            return True
        return False

    def __ne__(self, other) -> bool:
        return not other == self



from draft_sport.leagues.league import League
from draft_sport.leagues.pick import Pick
from draft_sport.leagues.team import Team
from draft_sport.leagues.invitation import Invitation

from draft_sport.fantasy.scores.player.score_card import ScoreCard
from draft_sport.fantasy.scores.player.order_by import OrderBy
from draft_sport.fantasy.scores.player.points import Points
from draft_sport.fantasy.scores.player.round import Round
from draft_sport.fantasy.scores.player.score import Score
from draft_sport.fantasy.player.profile import Profile
from draft_sport.fantasy.position.position import Position
from draft_sport.fantasy.position.constants import PositionConstants
from draft_sport.fantasy.season.season import Season
from draft_sport.fantasy.season.constants import SeasonConstants
from draft_sport.fantasy.competition.competition import Competition
from draft_sport.fantasy.competition.constants import CompetitionConstants
from draft_sport.fantasy.team.team import Team as FantasyTeam
from draft_sport.fantasy.team.constants import TeamConstants

from draft_sport.universe.position.category import PositionCategory

from draft_sport.data.order import Order

from draft_sport.security.session import Session
from draft_sport.security.perspective import Perspective
from draft_sport.security.secret_reset_request import SecretResetRequest

from draft_sport.humans.human import Human

from draft_sport.communication.challenge import Challenge
from draft_sport.communication.method import CommunicationMethod

from draft_sport.store.denomination import Denomination
from draft_sport.store.offer import Offer
from draft_sport.store.product import Product
from draft_sport.store.price import Price



"""
Draft Sport API
Secret Reset Request Module
author: hugh@blinkybeach.com
"""
from nozomi import Immutable, Decodable, NozomiTime, Configuration
from typing import Optional, Type, TypeVar, Any
from nozomi import URLParameter, URLParameters, HTTPMethod, ApiRequest

T = TypeVar('T', bound='SecretResetRequest')


class SecretResetRequest(Decodable):

    _PATH = '/human/secret-reset-request'

    def __init__(
        self,
        created: NozomiTime
    ) -> None:

        self._created = created

        return

    created = Immutable(lambda s: s._created)

    @classmethod
    def decode(cls: Type[T], data: Any) -> None:
        return cls(
            created=NozomiTime.decode(data['created'])
        )

    @classmethod
    def create(
        cls: Type[T],
        email_address: str,
        configuration: Configuration
    ) -> T:

        target = URLParameter('email', email_address)
        parameters = URLParameters([target])

        request = ApiRequest(
            path=cls._PATH,
            method=HTTPMethod.POST,
            configuration=configuration,
            url_parameters=parameters
        )

        return cls.decode(request.response_data)

    @classmethod
    def retrieve(
        cls: Type[T],
        reset_key: str,
        configuration: Configuration
    ) -> Optional[T]:

        target = URLParameter('reset-key', reset_key)
        parameters = URLParameters([target])

        request = ApiRequest(
            path=cls._PATH,
            method=HTTPMethod.GET,
            configuration=configuration,
            url_parameters=parameters
        )

        return cls.optionally_decode(request.response_data)



"""
Draft Sport Python
Position Category Module
author: hugh@blinkybeach.com
"""
from nozomi import Decodable, Immutable
from typing import TypeVar, Type, Any

T = TypeVar('T', bound='PositionCategory')


class PositionCategory(Decodable):

    def __init__(
        self,
        indexid: int,
        name: str
    ) -> None:

        self._indexid = indexid
        self._name = name

        return

    name = Immutable(lambda s: s._name)

    @classmethod
    def decode(cls: Type[T], data: Any) -> T:
        return cls(
            indexid=data['indexid'],
            name=data['name']
        )

    def __eq__(self, other) -> bool:
        if not isinstance(other, PositionCategory):
            return False
        if self._indexid == other._indexid:
            return True
        return False

    def __ne__(self, other) -> bool:
        return not other == self



"""
Draft Sport Python
Filled Composition Class
author: hugh@blinkbeach.com
"""
from typing import List
from nozomi import Immutable
from draft_sport.leagues.filled_requirement import FilledRequirement


class FilledComposition:
    """
    A utility class coralling Composition and Player data into a combined unit.
    Useful on the client side for presentation purposes. There is no
    equivalent object in the Draft Sport API
    """

    def __init__(
        self,
        requirements: List[FilledRequirement]
    ) -> None:

        self._requirements = requirements

        return

    requirements: List[FilledRequirement] = Immutable(lambda s: s._requirements)



"""
Draft Sport Python
Fantasy Competition Module
author: hugh@blinkybeach.com
"""
from draft_sport.fantasy.competition.competition import Competition
from nozomi import Decodable, HTTPMethod, URLParameter, URLParameters
from nozomi import Immutable, ApiRequest, RequestCredentials, Configuration
from nozomi import NozomiDate
from typing import TypeVar, Type, Any, Optional, List

T = TypeVar('T', bound='Season')


class Season(Decodable):

    _PATH = '/fantasy/season'
    _LIST_PATH = '/fantasy/season/list'

    def __init__(
        self,
        public_id: str,
        name: str,
        start_date: NozomiDate,
        competition: Competition
    ) -> None:

        self._public_id = public_id
        self._name = name
        self._start_date = start_date
        self._competition = competition

        return

    public_id = Immutable(lambda s: s._public_id)
    name = Immutable(lambda s: s._name)
    start_date = Immutable(lambda s: s._start_date)
    competition = Immutable(lambda s: s._competition)

    @classmethod
    def decode(cls: Type[T], data: Any) -> T:
        return cls(
            public_id=data['public_id'],
            name=data['name'],
            start_date=NozomiDate.decode(data['start_date']),
            competition=Competition.decode(data['competition'])
        )

    @classmethod
    def retrieve(
        cls: Type[T],
        public_id: str,
        configuration: Configuration,
        credentials: Optional[RequestCredentials] = None
    ) -> Optional[T]:

        if not isinstance(public_id, str):
            raise TypeError('public_id must be of type `str`')

        parameters = URLParameters([
            URLParameter('public_id', public_id)
        ])

        request = ApiRequest(
            path=cls._PATH,
            method=HTTPMethod.GET,
            data=None,
            url_parameters=parameters,
            credentials=credentials,
            configuration=configuration
        )

        return cls.optionally_decode(request.response_data)

    @classmethod
    def retrieve_list(
        cls: Type[T],
        configuration: Configuration,
        credentials: Optional[RequestCredentials] = None
    ) -> List[T]:

        request = ApiRequest(
            path=cls._LIST_PATH,
            method=HTTPMethod.GET,
            data=None,
            url_parameters=None,
            credentials=credentials,
            configuration=configuration
        )

        return cls.optionally_decode_many(
            request.response_data,
            default_to_empty_list=True
        )
